#!/usr/bin/env python2
# vim:fileencoding=utf-8
from __future__ import (unicode_literals, division, absolute_import,
                        print_function)

__license__ = 'GPL v3'
__copyright__ = '2018, 2019, 2020, 2021, 2022 Doitsu'

# standard libraries
import re, os, locale, sys, codecs, json
from os.path import expanduser, basename, isdir, isfile, join
from xml.sax.saxutils import escape

# Qt
from qt.core import (
    QTextEdit, QDockWidget, QApplication, QAction, 
    QFileDialog, QMessageBox, QDialog, QListWidget, QVBoxLayout, 
    QListWidgetItem, QDialogButtonBox, Qt, QEventLoop, QBrush, QColor,
    QStyle, QStyleOptionViewItem, QStyledItemDelegate, QPalette, 
    QAbstractTextDocumentLayout, QTextDocument, QSize
)

# Calibre libraries
from calibre.gui2 import error_dialog, info_dialog
from calibre.gui2.tweak_book.plugin import Tool
from calibre.gui2.tweak_book import editors, editor_name
from calibre.gui2.tweak_book.file_list import FileList
from calibre.ebooks.oeb.polish.container import OEB_DOCS
from calibre.utils.config import JSONConfig
from calibre.utils.localization import get_lang
from calibre.utils.config import config_dir
from calibre.constants import iswindows, islinux, isosx

# make sure the plugin will work with the Python 3 version of Calibre
if sys.version_info[0] == 2:
    from urllib import urlopen
else:
    from urllib.request import urlopen

#---------------------------------
# jar wrapper for LanguageTool
#---------------------------------
def jarWrapper(*args):
    ''' runs the LT jar file '''
    import subprocess
    startupinfo = None

    # stop the windows console popping up every time the prog is run
    if iswindows:
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = subprocess.SW_HIDE

    process = subprocess.Popen(list(args), stdout=subprocess.PIPE, stderr=subprocess.PIPE, startupinfo=startupinfo)
    ret = process.communicate()
    returncode = process.returncode
    return ret, returncode

#-------------------------------------
# remove all html tags and entities
#-------------------------------------
def DeleteHTMLTags(html):
    ''' removes all HTML tags and entities '''
    zwj = u'\u200d'
    cleaned_html = html

    # replace XML declaration
    match = re.search(r'<\?xml[^>]+>', cleaned_html)
    if match is not None:
        match_found = match.group(0)
        replacement = zwj * len(match_found)
        cleaned_html = cleaned_html.replace(match_found, replacement)

    # replace doctype
    match = re.search(r'<!DOCTYPE[^>]+>', cleaned_html)
    if match is not None:
        match_found = match.group(0)
        if not '\n' in match_found:
            replacement = zwj * len(match_found)
        else:
            string_list = list(match_found)
            for letter in range(0, len(match_found)):
                if string_list[letter] not in ['\r', '\n', ' ']:
                    string_list[letter] = zwj
            replacement = "".join(string_list)
        cleaned_html = cleaned_html.replace(match_found, replacement)

    # replace script sections
    match = re.search('<script[^>]*>.*</script>', cleaned_html, re.MULTILINE|re.DOTALL)
    if match is not None:
        match_found = match.group(0)
        if not '\n' in match_found:
            replacement = zwj * len(match_found)
        else:
            string_list = list(match_found)
            for letter in range(0, len(match_found)):
                if string_list[letter] not in ['\r', '\n', ' ']:
                    string_list[letter] = zwj
            replacement = "".join(string_list)
        cleaned_html = cleaned_html.replace(match_found, replacement)

    # replace standard HTML tags
    matches = re.finditer(r'''</?([A-Za-z][^\s>/]*)(?:=\s*(?:"[^"]*"|'[^']*'|[^\s>]+)|[^>])*(?:>|$)''', cleaned_html)
    for match in matches:
        match_found = match.group(0)
        replacement = ' ' * len(match_found)
        cleaned_html = cleaned_html.replace(match_found, replacement)

    # replace entities
    matches = re.finditer(r'&[^;]+;', cleaned_html)
    for match in matches:
        match_found = match.group(0)
        # non-breaking space
        if match_found in ['&#160;', '&nbsp;', '&#xa0;']:
            replacement = u'\u200D' * (len(match_found) -1) + u'\u00A0'
        # en dash
        elif match_found in ['&#8211;', '&ndash;', '&#x2013;']:
            replacement = u'\u200D' * int(len(match_found) / 2) + u'\u2013' + u'\u200D' * 3
        # em dash
        elif match_found in ['&#8212;', '&mdash;', '&#x2014;']:
            replacement = u'\u200D' * int(len(match_found) / 2) + u'\u2014' + u'\u200D' * 3
        # all other entities
        else:
            replacement = ' ' * len(match_found)
        cleaned_html = cleaned_html.replace(match_found, replacement)

    return cleaned_html

#-------------------------------------------------------------------------------------
# temp folder; https://www.mobileread.com/forums/showpost.php?p=3233865&postcount=4
#-------------------------------------------------------------------------------------
from contextlib import contextmanager

@contextmanager
def make_temp_directory():
    ''' creates a temp folder '''
    import tempfile
    import shutil
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)

#-------------------------------------------------------------------------------------------------
# delegate for displaying colored text; https://stackoverflow.com/a/53603442/2614117
#-------------------------------------------------------------------------------------------------
class HTMLDelegate(QStyledItemDelegate):
    ''' a delegate for displaying colored text '''
    def __init__(self, parent=None):
        super(HTMLDelegate, self).__init__(parent)
        self.doc = QTextDocument(self)

    def paint(self, painter, option, index):
        painter.save()
        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        self.doc.setHtml(options.text)
        options.text = ""
        style = QApplication.style() if options.widget is None \
            else options.widget.style()
        style.drawControl(QStyle.ControlElement.CE_ItemViewItem, options, painter)

        ctx = QAbstractTextDocumentLayout.PaintContext()
        if option.state & QStyle.StateFlag.State_Selected:
            ctx.palette.setColor(QPalette.Text, option.palette.color(
                QPalette.Active, QPalette.HighlightedText))
        else:
            ctx.palette.setColor(QPalette.Text, option.palette.color(
                QPalette.Active, QPalette.Text))
        textRect = style.subElementRect(QStyle.SubElement.SE_ItemViewItemText, options, None)
        if index.column() != 0:
            textRect.adjust(5, 0, 0, 0)
        constant = 4
        margin = (option.rect.height() - options.fontMetrics.height()) // 2
        margin = margin - constant
        textRect.setTop(textRect.top() + margin)

        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        self.doc.documentLayout().draw(painter, ctx)
        painter.restore()
        s = QSize(int(self.doc.idealWidth()), int(self.doc.size().height()))
        index.model().setData(index, s, Qt.SizeHintRole)

#============================
# plugin class
#============================

class LTTool(Tool):
    #: a unique name that will be used as a key
    name = 'language-tool'

    #: If True the user can choose to place this tool in the plugins toolbar
    allowed_in_toolbar = True

    #: If True the user can choose to place this tool in the plugins menu
    allowed_in_menu = True

    def create_action(self, for_toolbar=True):
        # Create an action, this will be added to the plugins toolbar and
        # the plugins menu
        ac = QAction(get_icons('images/icon.png'), 'Run LanguageTool', self.gui)  # noqa
        if not for_toolbar:
            # Register a keyboard shortcut for this toolbar action.
            self.register_shortcut(ac, 'language-tool', default_keys=('Ctrl+Shift+Alt+Z',))
        ac.triggered.connect(self.run_lt)
        return ac

    #==============================
    # the main routine
    #==============================
    def run_lt(self):

        # create a savepoint
        self.boss.add_savepoint('Before: LanguageTool')

        # commit changes in open editor windows
        self.boss.commit_all_editors_to_container()

        # get the Calibre GUI language
        calibre_ui_lang = get_lang()

        # get current container
        container = self.current_container

        if not container:
            return info_dialog(self.gui, 'No book open', 'Need to have a book open first.', show=True)

        #----------------------------------
        # get ebook language metadata
        #----------------------------------
        orig_lang = 'en'
        dc_language = 'en'
        try:
            # I couldn't use container.mi.languages[0], because it
            # returns normalized language codes. E.g. en-US => eng
            orig_lang = container.opf_xpath('//dc:language/text()')[0]
            if orig_lang != '':
                dc_language = orig_lang
                if len(dc_language) == 2 and dc_language.isupper(): dc_language = dc_language.lower()
        except:
            pass

        #--------------------------------------------------------------------------------------------------
        # change two letter language codes to four letter language codes for better LT results
        #--------------------------------------------------------------------------------------------------
        lang_dict = {
            'ast' : 'ast-ES', 'be' : 'be-BY', 'br' : 'br-FR' , 'da' : 'da-DK' , 'ga' : 'ga-IE' ,
            'el' : 'el-GR', 'gl' : 'gl-ES', 'ja' : 'ja-JP', 'km' : 'km-KH', 'pl' : 'pl-PL', 'ro' : 'ro-RO',
            'ru' : 'ru-RU', 'sk' : 'sk-SK', 'sl' : 'sl-SI', 'ta' : 'ta-IN', 'tl' : 'tl-PH', 'uk' : 'uk-UA',
            'en' : 'en-US'
        }
        if dc_language in lang_dict:
            dc_language = lang_dict[dc_language]

        #-------------------------------------------------------------
        # make sure that the language is actually supported by LT
        #-------------------------------------------------------------
        suported_language_codes = ['ar', 'ast-ES', 'be-BY', 'br-FR', 'ca-ES', 'ca-ES-balear',  'ca-ES-valencia', 
            'crh-UA', 'da-DK', 'de', 'de-AT', 'de-CH', 'de-DE', 'de-DE-x-simple-language', 'el-GR', 'en', 'en-AU', 
            'en-CA', 'en-GB', 'en-NZ', 'en-US', 'en-ZA', 'eo', 'es', 'es-AR',  'fa', 'fr', 'fr-BE', 'fr-CA', 'fr-CH', 'ga-IE'  
            'gl-ES', 'it', 'ja-JP', 'km-KH', 'nl', 'nl-BE', 'pl-PL', 'pt', 'pt-AO', 'pt-BR', 'pt-MZ', 'pt-PT', 'ro-RO', 'ru-RU', 
            'sk-SK', 'sl-SI', 'sr', 'sr-BA', 'sr-HR', 'sr-ME', 'sr-RS', 'sv', 'ta-IN', 'tl-PH', 'uk-UA', 'zh-CN']

        if not dc_language in suported_language_codes:
            dc_language = dc_language[0:2]
            if not dc_language in suported_language_codes:
                return error_dialog(self.gui, 'Language code not supported!', "'{}' is not a language code known to LanguageTool. Supported language codes are: {}.".format(orig_lang, ', '.join(suported_language_codes)), show=True)

        #=================================
        # set default preferences
        #=================================

        # select language-specific rules to disable spell check
        if calibre_ui_lang == 'fr':
            default_rules = 'FR_SPELLING_RULE'
        elif calibre_ui_lang == 'it':
            default_rules = 'MORFOLOGIK_RULE_IT_IT'
        elif calibre_ui_lang == 'de':
            default_rules = 'GERMAN_SPELLER_RULE,GERMAN_WORD_REPEAT_BEGINNING_RULE'
            # OLD_SPELLING,OLD_SPELLING_INTERNAL,EMPFOHLENE_RECHTSCHREIBUNG
            # "unreformed" German = de-1901
        elif calibre_ui_lang == 'es':
            default_rules = 'MORFOLOGIK_RULE_ES'
        elif calibre_ui_lang == 'nl':
            default_rules = 'MORFOLOGIK_RULE_NL_NL'
        elif calibre_ui_lang == 'pl':
            default_rules = 'MORFOLOGIK_RULE_PL_PL'
        elif calibre_ui_lang == 'ca':
            default_rules = 'MORFOLOGIK_RULE_CA_ES'
        elif calibre_ui_lang == 'ru':
            default_rules = 'MORFOLOGIK_RULE_RU_RU'
        else:
            default_rules = 'MORFOLOGIK_RULE_EN_US,MORFOLOGIK_RULE_EN_GB,ENGLISH_WORD_REPEAT_BEGINNING_RULE'
            # MORFOLOGIK_RULE_EN_CA,MORFOLOGIK_RULE_EN_ZA,MORFOLOGIK_RULE_EN_AU,MORFOLOGIK_RULE_EN_NZ

        # disable all white space rules (returns false positives, because tags have been replaced with spaces)
        default_rules += ',WHITESPACE_RULE,COMMA_PARENTHESIS_WHITESPACE,SENTENCE_WHITESPACE'

        # save default preferences
        prefs = JSONConfig('plugins/LanguageTool')
        if prefs == {}:
            prefs['update_check'] = False
            prefs['allFiles'] = False
            prefs['close_cb'] = False
            prefs['clipboard_copy'] = False
            prefs['ngramIndexDir'] = False
            prefs['disabledRules'] = default_rules
            prefs['disabledCategories'] = 'REDUNDANCY'
            prefs.commit()

        #----------------------------------------
        # get preferences
        #----------------------------------------

        # used for debugging
        debug = prefs.get('debug', False)

        # if True, results will be copied to the clipboard
        clipboard_copy = prefs.get('clipboard_copy', False)

        # if True, the plugin will check for new LT versions
        update_check = prefs.get('update_check', False)

        # if True, all files will be checked
        allFiles = prefs.get('allFiles', False)

        # if True, the Check Book panel will be automatically closed
        close_cb = prefs.get('close_cb', False)

        # the path to languagetool-commandline.jar
        ltPath = prefs.get('ltPath', None)

        # the directory that cointains the ngram files (ngram checking is very slow!)
        ngramIndexDir = prefs.get('ngramIndexDir', None)

        # IDs of rules to be enabled, comma-separated (-e)
        enabledRules = prefs.get('enabledRules', None)

        # IDs of rules to be disabled, comma-separated (-d)
        disabledRules = prefs.get('disabledRules', default_rules)
        disabledRulesList = []
        if disabledRules != '':
            disabledRulesList = disabledRules.split(',')
        enabledCategories = prefs.get('enabledCategories', None)

        # IDs of categories to be enabled, comma-separated (--enablecategories)
        disabledCategories = prefs.get('disabledCategories', 'REDUNDANCY')
        disabledCategoriesList = []
        if disabledCategories != '':
            disabledCategoriesList = disabledCategories.split(',')

        # If true, only the rules and categories whose IDs are specified with
        # enabledRules and/or enabledCategories are enabled.
        enabledOnly = prefs.get('enabledOnly', None)

        #---------------------------------------------------
        # get/verify languagetool-commandline.jar path
        #---------------------------------------------------
        if not ltPath or not isfile(ltPath):
            home = expanduser('~')
            file_name, filter = QFileDialog.getOpenFileName(self.gui, 'Select LanguageTool Java file', home, 'LT (languagetool-commandline.jar)')
            if basename(file_name) == 'languagetool-commandline.jar':
                ltPath = file_name
                prefs['ltPath'] = ltPath
                prefs.commit()
            else:
                return error_dialog(self.gui, 'LT not selected!', 'You didn\'t select languagetool-commandline.jar!', show=True)

        #----------------------------------------
        # define LT command line parameters
        #----------------------------------------
        args = ['java', '-Dfile.encoding=UTF8', '-Xmx1024m', '-Xms512m', '-jar', ltPath, '-c', 'utf-8']

        # path to the ngram-index directory (--languagemodel)
        if ngramIndexDir and not isdir(ngramIndexDir):
            ngramIndexDir = None

        if ngramIndexDir:
            args.extend(['--languagemodel', ngramIndexDir])

        # IDs of rules to be enabled, comma-separated (-e)
        if not enabledRules and ngramIndexDir:
            args.extend(['-e', 'CONFUSION_RULE'])

        if enabledRules and not ngramIndexDir:
            args.extend(['-e', enabledRules.replace(' ', '')])

        # IDs of rules to be disabled, comma-separated (-d)
        if disabledRules and not ngramIndexDir:
            args.extend(['-d', disabledRules.replace(' ', '')])

        # IDs of categories to be enabled, comma-separated (--enablecategories)
        if enabledCategories and not ngramIndexDir:
            args.extend(['--enablecategories', enabledCategories.replace(' ', '')])

        # IDs of categories to be disabled, comma-separated to be disabled (--disablecategories)
        if disabledCategories and not ngramIndexDir:
            args.extend(['--disablecategories', disabledCategories.replace(' ', '')])

        # If true, only the rules and categories whose IDs are specified with
        # enabledRules or enabledCategories are enabled. (--enabledonly)
        if (ngramIndexDir and not enabledOnly) or enabledOnly:
            args.extend(['-eo'])

        # check for optional userrules.xml file
        userRulesPath = join(config_dir, 'plugins', 'user-rules.xml')
        if isfile(userRulesPath) and not ngramIndexDir:
            args.extend(['--rulefile', userRulesPath])

        args.extend(['--json', '-l', dc_language])

        #----------------------------------------
        # create a temp folder
        #----------------------------------------
        with make_temp_directory() as new_temp_dir:

            error_messages = []
            if clipboard_copy and not debug:
                clipboard_messages = '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format('File Name', 'Line', 'Column', 'Context', 'Catgory', 'Rule', 'Message')

            # get all OPEN editor windows
            #all_editors = []
            #for name in editors.keys():
                #all_editors.append((name, None))

            # get ACTIVE editor file
            active_editor = []
            name = editor_name(self.boss.gui.central.current_editor)
            if name and container.mime_map[name] in OEB_DOCS:
                active_editor.append((name, None))

            # get ALL files in the spine
            all_files = list(container.spine_names)

            # select the file(s) to be processed
            if allFiles:
                file_list = all_files
            else:
                if active_editor != []:
                    file_list = active_editor
                else:
                    return info_dialog(self.gui, 'No open HTML file found.', 'Please open an HTML file, before running the plugin.', show=True)

            # display busy cursor
            QApplication.setOverrideCursor(Qt.WaitCursor)

            #--------------------------
            # process files
            #--------------------------
            for name, linear in file_list:

                # get raw html & plain text
                raw_html = container.raw_data(name)
                plain_text = DeleteHTMLTags(raw_html)

                # write plain text to temporary text file
                file_name = join(new_temp_dir, basename(name))
                with codecs.open(file_name, 'w', encoding='utf-8') as f:
                    f.write(plain_text)

                #--------------------------------------------------------------------
                # run LanguageTool
                #-------------------------------------------------------------------
                run_args = []
                run_args.extend(args)
                run_args.append(file_name)
                result, returncode = jarWrapper(*run_args)
                stdout = result[0].decode('utf-8')
                stderr = result[1].decode('utf-8')

                #================================
                # parse JSON output
                #================================
                if returncode == 0:

                    # define color tags; #DC143C = Crimson
                    start_color = ' <b><span style="color: #DC143C;">'
                    end_color = '</span></b> '

                    # process LT JSON
                    parsed_json = (json.loads(stdout))

                    # get software name & version
                    software_name = parsed_json['software']['name']
                    software_version = parsed_json['software']['version']
                    LT_version = 'v' + software_version

                    # get error messages
                    matches = parsed_json['matches']
                    print(len(matches), 'error(s) found.')

                    # parse error messages
                    for match in matches:

                        # get message & short message
                        message = match['message']
                        try:
                            shortmsg = match['shortMessage']
                        except:
                            shortmsg = ''

                        # merge error messages
                        if 'shortmsg' != '':
                            error_msg = shortmsg + ': ' + message
                        else:
                            error_msg = message

                        # get & format suggestions
                        replacements = match['replacements']
                        if len(replacements) == 0:
                            html_suggestions = ''
                            plain_suggestions = ''
                        elif len(replacements) == 1:
                            html_suggestions = ' Suggestion: ' + start_color + escape(replacements[0]['value']) + end_color.strip()
                            plain_suggestions = ' Suggestion: ' + '**' + replacements[0]['value'] + '**'
                        else:
                            html_suggestions = ''
                            plain_suggestions = ''
                            for replacement in replacements:
                                html_suggestions += start_color + escape(replacement['value']) + end_color.strip() + ', '
                                plain_suggestions += '**' + escape(replacement['value']) + '**' + ', '
                            html_suggestions = ' Suggestions: ' + html_suggestions[0:len(html_suggestions)-2]
                            plain_suggestions = ' Suggestions: ' + plain_suggestions[0:len(plain_suggestions)-2]

                        # get error offset and length (error_length = context_length)
                        error_offset = match['offset']
                        # error_length = match['length']

                        # get line number
                        line_number = plain_text.count('\n', 0, error_offset) + 1

                        # get column number
                        column_number = error_offset - plain_text.rfind('\n', 0, error_offset)

                        # format context
                        context_text = match['context']['text']
                        context_offset = match['context']['offset']
                        context_length = match['context']['length']
                        typo = context_text[context_offset:context_offset+context_length].strip()
                        plain_typo = ' **' + typo + '** '
                        typo = escape(typo).replace('"', "&quot;")
                        html_typo = start_color + typo + end_color

                        # format context
                        textbefore = context_text[0:context_offset].strip()
                        textafter = context_text[context_offset + context_length:].strip()
                        plain_context = re.sub('[\t ]{2,}', ' ', textbefore + plain_typo + textafter)
                        textbefore = escape(textbefore).replace('"', "&quot;")
                        textafter = escape(textafter).replace('"', "&quot;")
                        html_context = re.sub('[\t ]{2,}', ' ', textbefore + html_typo + textafter)

                        rule_id = match['rule']['id']
                        rule_description = match['rule']['description']
                        rule_category_id = match['rule']['category']['id']
                        #rule_category_name = match['rule']['category']['name']

                        # format location
                        location = basename(name) + ' Line: ' + str(line_number) + ', Col: ' + str(column_number)

                        # format message
                        message = '{}. Context: {} <b>{}:{}</b> {}{}'.format(location, html_context, rule_category_id, rule_id, error_msg, html_suggestions)

                        # LanguageTool sometimes ignores disabled internal rules and categories
                        if rule_id not in disabledRulesList and rule_category_id not in disabledCategoriesList:
                            error_messages.append((name, line_number, column_number, message))
                            if clipboard_copy and not debug:
                                clipboard_messages += '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(basename(name), line_number, column_number, plain_context, rule_category_id, rule_id, error_msg + plain_suggestions)

                else:
                    # LT failed; display command line params and error messages
                    QApplication.restoreOverrideCursor()
                    return error_dialog(self.gui, 'LT error', ' '.join(run_args) + '\n' + stdout + '\n' + stderr, show=True)

        #=================================
        # display LT results
        #=================================

        if debug:
            # copy raw LT output to clipboard
            QApplication.clipboard().setText(' '.join(run_args) + '\n' + stdout + '\n' + stderr)
            #QApplication.clipboard().setText(str(error_messages))
            #QApplication.clipboard().setText(plain_text)

        if error_messages != []:

            # copy LT messages to the system clipboard
            if clipboard_copy and not debug:
                QApplication.clipboard().setText(clipboard_messages)

            #---------------------------------------------------------------
            # auxiliary routine for loading the file into the editor
            #---------------------------------------------------------------
            def GotoLine(item):

                # get list item number
                current_row = listWidget.currentRow()

                # get error information
                filepath, line, col, message = error_messages[current_row]

                # go to the file
                self.boss.edit_file(filepath)
                editor = self.boss.gui.central.current_editor
                if editor is not None and editor.has_line_numbers and line is not None:
                    if col is not None:
                        editor.editor.go_to_line(int(line), col=int(col) - 1)
                    else:
                        editor.current_line = int(line)
                else:
                    error_dialog(self.gui, 'Internal error', 'Editor or line number not found.', show=True)

            #------------------------------------------------------------
            # remove existing LT dock and close Check Ebook dock
            #------------------------------------------------------------
            for widget in self.gui.children():
                if isinstance(widget, QDockWidget) and widget.objectName() == 'lt-dock':
                    #self.gui.removeDockWidget(widget)
                    #widget.close()
                    widget.setParent(None)
                if isinstance(widget, QDockWidget) and widget.objectName() == 'check-book-dock' and close_cb == True:
                    widget.close()

            #----------------------------------
            # define list widget layout
            #----------------------------------
            listWidget = QListWidget()
            list_palette = QPalette()
            # set custom highlight color PapayaWhip (#ffefd5 )
            list_palette.setColor(QPalette.Highlight, QColor(255, 239, 213))
            list_palette.setColor(QPalette.HighlightedText, Qt.black)
            listWidget.setPalette(list_palette)
            delegate = HTMLDelegate(listWidget)
            listWidget.setItemDelegate(delegate)
            l = QVBoxLayout()
            l.addWidget(listWidget)

            #----------------------------------
            # define dock widget layout
            #----------------------------------
            dock_widget = QDockWidget(self.gui)
            dock_widget.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea | Qt.BottomDockWidgetArea | Qt.TopDockWidgetArea)
            dock_widget.setObjectName('lt-dock')
            dock_widget.setWindowTitle('LanguageTool')
            dock_widget.setWidget(listWidget)

            #--------------------------------------------
            # add error messages to list widget
            #--------------------------------------------
            for error_msg in error_messages:
                filename, line, col, message = error_msg
                item = QListWidgetItem(message)
                listWidget.addItem(item)

            # connect click to action
            listWidget.itemClicked.connect(GotoLine)

            # add dock widget to the dock
            self.gui.addDockWidget(Qt.TopDockWidgetArea, dock_widget)
            QApplication.restoreOverrideCursor()
        else:
            QApplication.restoreOverrideCursor()
            # remove existing LT dock
            for widget in self.gui.children():
                if isinstance(widget, QDockWidget) and widget.objectName() == 'lt-dock':
                    widget.setParent(None)

            # display no errors found message
            info_dialog(self.gui, 'No errors found.', 'LanguageTool didn\'t find any errors.', show=True)

        #-----------------------------------------------------------------
        # check for LT updates (no check for SNAPSHOT versions)
        #-----------------------------------------------------------------
        if update_check and 'SNAPSHOT' not in LT_version:
            try:
                github_url = 'https://api.github.com/repos/languagetool-org/languagetool/tags'
                response = urlopen(github_url).read().decode('utf-8')

                parsed_json = json.loads(response)
                latest_version = parsed_json[1]['name']
                if latest_version != LT_version:
                    return info_dialog(self.gui, 'New LT Version', 'New LangugeTool version available: ' + latest_version, show=True)
            except:
                pass